import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;

// TODO: Auto-generated Javadoc
/**
 * The Class CodeGenerator.
 */
public class CodeGenerator {
	
	/** Hashmap tabela simbolos global. */
	private HashMap<String,String> tabelaSimbolosGlobal;
	
	/** ArrayList functions. */
	private ArrayList<StructFunction> functions = new ArrayList<StructFunction >();
	
	/** ArrayList globais. */
	private ArrayList<GlobalVariables> globais = new ArrayList<GlobalVariables>();
	
	/** Hashmap variaveis func. */
	private HashMap<String,Integer> variaveisFunc;
	
	/** The filename. */
	private String filename;
	
	/** Simplenode node. */
	private SimpleNode node;
	
	/** String nome modulo. */
	private String nomeModulo="";
	
	/** File fich1. */
	private File fich1;
	
	/** The max loops. */
	private int nLabel = 0, nLabels=0, maxLabel=0, nLoops=0, maxLoops=0;
	
	/**
	 * Instantiates a new code generator.
	 *
	 * @param String the name
	 * @param HashMap the globals
	 * @param ArrayList the functions
	 * @param File the output file
	 * @param SimpleNode the node
	 */
	CodeGenerator(String name, HashMap<String,String > globals, ArrayList<StructFunction > func, File fich1, SimpleNode node)
	{
		this.tabelaSimbolosGlobal=globals;
		this.functions=func;
		this.filename=name;
		this.node=node;	
		this.fich1=fich1;
	}
	
	/**
	 * Function that creates the code generation and invokes Jasmin
	 *
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	void create() throws IOException
	{
		String fName = filename.substring(0,filename.length()-4)+".j";
		System.out.println("\nNome do ficheiro a criar: " + fName.substring(0, fName.length()-2));
		FileOutputStream fich = new FileOutputStream(fName);
		fich.flush();

		constroiGlobais(node,fich);		
		constroiFuncoes(node,fich);
	
		fich.close();
			
		//execucao do jasmin
		String s="";
	    Process p = Runtime.getRuntime().exec("java -jar jasmin.jar " + fName);
	    
	    BufferedReader stdInput = new BufferedReader(new InputStreamReader(p.getInputStream()));
	    BufferedReader stdError = new BufferedReader(new InputStreamReader(p.getErrorStream()));

	    while ((s = stdError.readLine()) != null) {
            System.out.println(s);
            System.exit(0);
        }
        
        while ((s = stdInput.readLine()) != null) {
            System.out.println(s);
        }   
        stdInput.close();
        stdError.close();
        fich.close();
        fich1.delete();        
 }
	
	/**
	 * Constroi funcoes.
	 *
	 * @param SimpleNode the n
	 * @param FileOutputStream the fich
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	void constroiFuncoes(SimpleNode n, FileOutputStream fich) throws IOException
	{
		for(int j=0; j < functions.size(); j++)
		{
			variaveisFunc = new HashMap<String,Integer>();
			for(int i=0; i < n.children.length; i++)
			{
				if(n.children[i].toString().equals("FUNCTION")) 
				{
					String nFunc="",s2="", s3="";
					s2=((SimpleNode) (n.children[i]).jjtGetChild(0).jjtGetChild(0)).nameF;
					s3=((SimpleNode) (n.children[i]).jjtGetChild(1)).nameF;
					
					Object[] keys = functions.get(j).args.keySet().toArray();
					
					nFunc=s3;
					try{
						if(s3.equals(""));						
					}catch (java.lang.NullPointerException e)
					{
						nFunc=s2;
					}					
					
					if(functions.get(j).name.equals(nFunc))
					{
						if(functions.get(j).name.equals("main"))
							variaveisFunc.put("StringJava",0);		

						for(int w= 0; w < keys.length; w++)
							if(!variaveisFunc.containsKey((String) keys[w]))
								variaveisFunc.put((String) keys[w], functions.get(j).args_order.get(keys[w]));
						
						if(!functions.get(j).type_return.equals("void"))
							if(!variaveisFunc.containsKey(functions.get(j).var_ret))
								variaveisFunc.put(functions.get(j).var_ret, variaveisFunc.size());
						
						String s1 = "\n.method public static " + functions.get(j).name + "(";
						
						if(functions.get(j).name.equals("main"))
							s1+="[Ljava/lang/String;" ;
						else
						{
							Object[] keys1 = functions.get(j).args.keySet().toArray();
							for(int i1=0; i1 < keys1.length; i1++)
							{
								if(functions.get(j).args.get(keys[i1]).equals("Scalar"))
									s1+="I";
								else
									s1+="[I";
							}				
						}

						if(functions.get(j).type_return.equals("void"))
							s1+= ")V";
						else if(functions.get(j).type_return.equals("Scalar"))
							s1+= ")I";
						else if(functions.get(j).type_return.equals("ScalarArray"))
							s1+= ")[I";
						
						int numLocals=0, n11=0;
						Object[] keys2 = functions.get(j).args.keySet().toArray();
						for(int h=0; h < keys2.length; h++)
						{
							if(functions.get(j).var_local.containsKey(keys2[h]))
								n11++;
						}
						numLocals = functions.get(j).args.size() + functions.get(j).var_local.size() - n11 + 1;
						
						s1+="\n.limit stack 30\n.limit locals " + numLocals + "\n";
						fich.write(s1.getBytes());
						
						SimpleNode n1 = (SimpleNode) n.children[i];
						for(int l=0; l < n1.children.length; l++)
						{	
							if(n1.children[l].toString().equals("STMLIST")) 
							{
								SimpleNode n2 = (SimpleNode) n1.children[l];
								processStmList(n2,j,fich);
							}
						}												
					}					
				}
			}
			
			
			//atribuir variavel de retorno da funcao						
			if(!functions.get(j).type_return.equals("void"))
			{
				String var_ret = functions.get(j).var_ret, ret="";
				
				if(variaveisFunc.containsKey(var_ret) && functions.get(j).type_return.equals("Scalar"))
				{
					if(variaveisFunc.get(var_ret)>= 0 && variaveisFunc.get(var_ret)<=3 )
						ret = "\tiload_" + variaveisFunc.get(var_ret) + "\n";
					else
						ret = "\tiload " + variaveisFunc.get(var_ret) + "\n";
				}
				else if(variaveisFunc.containsKey(var_ret) && functions.get(j).type_return.equals("ScalarArray"))
				{
					if(variaveisFunc.get(var_ret) >= 0 && variaveisFunc.get(var_ret) <= 3)
						ret = "\taload_" + variaveisFunc.get(var_ret) + "\n";
					else
						ret = "\taload " + variaveisFunc.get(var_ret) + "\n";
				}
				if(this.tabelaSimbolosGlobal.containsKey(var_ret))
				{
					String tmp1="";
					ret = "getstatic " + nomeModulo + "/" + var_ret + " ";
					if(functions.get(j).var_local.containsKey(var_ret))
						tmp1= functions.get(j).var_local.get(var_ret);
					else if(functions.get(j).args.containsKey(var_ret))
						tmp1= functions.get(j).args.get(var_ret);
					else if(tabelaSimbolosGlobal.containsKey(var_ret))
						tmp1= tabelaSimbolosGlobal.get(var_ret);
					
					if(tmp1.equals("Scalar"))
						ret += "I";
					else
						ret +="[I";
					
					ret+="\n";
				}			
				fich.write(ret.getBytes());

				//ireturn ou areturn
				if(functions.get(j).type_return.equals("Scalar"))
					fich.write("\tireturn\n".getBytes());
				else
					fich.write("\tareturn\n".getBytes());
			}
			
			String s5 = "return\n.end method\n";
			fich.write(s5.getBytes());
		}	
		
	}

	/**
	 * Process Statement List.
	 *
	 * @param n2 the n2
	 * @param j the j
	 * @param fich the fich
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	void processStmList (SimpleNode n2, int j, FileOutputStream fich) throws IOException
	{
		//analisar conteudo da funcao
		
		for(int k=0; k < n2.children.length; k++)	//verifica cada statement
		{
			SimpleNode n3 = (SimpleNode) n2.children[k];
			for(int p=0; p < n3.children.length; p++)
			{
				if(n3.children[p].toString().equals("ASSIGN"))
					processAssign(n3,j, fich);
				else if(n3.children[p].toString().equals("WHILE"))				
					processWhile(n3,j, fich, maxLoops - nLoops + 1);					
				else if(n3.children[p].toString().equals("CALL"))
					processCall(n3, j, fich);										
				else if(n3.children[p].toString().equals("IF"))
					processIf(n3,j, fich, maxLabel + nLabels*2+1);
			}	
		}
	}
		
	/**
	 * Process while.
	 *
	 * @param SimpleNode the n3
	 * @param Integer the j
	 * @param FileOutputStream the fich
	 * @param Integer the tmp
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	private void processWhile(SimpleNode n3, int j, FileOutputStream fich, int tmp) throws IOException {
		
		nLoops += tmp;
		
		if(nLoops > maxLoops)
			maxLoops=nLoops;
		
		fich.write(("loop" + nLoops + ":\n").getBytes());
		
		SimpleNode ExpressionTest = (SimpleNode)n3.jjtGetChild(0).jjtGetChild(1);		
		SimpleNode LHS = (SimpleNode)ExpressionTest.jjtGetChild(0);		
		processLHS(LHS,j,fich, 1);		
		SimpleNode Oper = (SimpleNode)ExpressionTest.jjtGetChild(1);	
		SimpleNode RHS = (SimpleNode)ExpressionTest.jjtGetChild(2);
		processRHS(RHS,j,fich);
		
		String oper="", s= "loop_end";
		
		if(Oper.operation.toString().equals("=="))
			oper+="\tif_icmpne " + s +  nLoops + "\n";
		else if(Oper.operation.toString().equals(">="))
			oper+="\tif_icmplt " + s +  nLoops + "\n";
		else if(Oper.operation.toString().equals(">"))
			oper+="\tif_icmple " + s +  nLoops + "\n";
		else if(Oper.operation.toString().equals("<="))
			oper+="\tif_icmpgt " + s +  nLoops + "\n";
		else if(Oper.operation.toString().equals("<"))
			oper+="\tif_icmpge " + s +  nLoops + "\n";
		else if(Oper.operation.toString().equals("!="))
			oper+="\tif_icmpeq " + s +  nLoops + "\n";		
		fich.write(oper.getBytes());
				
		SimpleNode StmLst = (SimpleNode)n3.jjtGetChild(0).jjtGetChild(2);
		nLoops++;
		processStmList((SimpleNode) StmLst, j, fich);
		nLoops--;
		
		fich.write(("\tgoto loop"+ nLoops +"\nloop_end" + nLoops +":\n").getBytes());		
	}

	/**
	 * Process if.
	 *
	 * @param SimpleNode the n3
	 * @param Integer the j
	 * @param FileOutputStream the fich
	 * @param Integer the l
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	private void processIf(SimpleNode n3, int j, FileOutputStream fich, int l) throws IOException {
		
		nLabel=l;
			
		if(nLabel > maxLabel)
			maxLabel=nLabel;		
		
		SimpleNode ExpressionTest = (SimpleNode)n3.jjtGetChild(0).jjtGetChild(1);
		SimpleNode LHS = (SimpleNode)ExpressionTest.jjtGetChild(0);
		processLHS(LHS,j,fich, 1);
		SimpleNode Oper = (SimpleNode)ExpressionTest.jjtGetChild(1);
		SimpleNode RHS = (SimpleNode)ExpressionTest.jjtGetChild(2);
		processRHS(RHS,j,fich);		
		SimpleNode StmLst = (SimpleNode)n3.jjtGetChild(0).jjtGetChild(2);
		String oper="",label = "Label" + nLabel;
		
		if(Oper.operation.toString().equals("=="))
			oper+="\tif_icmpeq " + label + "\n";
		else if(Oper.operation.toString().equals(">="))
			oper+="\tif_icmpge " + label + "\n";
		else if(Oper.operation.toString().equals(">"))
			oper+="\tif_icmpgt " + label + "\n";
		else if(Oper.operation.toString().equals("<="))
			oper+="\tif_icmple " + label + "\n";
		else if(Oper.operation.toString().equals("<"))
			oper+="\tif_icmplt " + label + "\n";
		else if(Oper.operation.toString().equals("!="))
			oper+="\tif_icmpne " + label + "\n";		
		fich.write(oper.getBytes());
	
		if(((SimpleNode)n3.jjtGetChild(0)).children.length == 5)
		{
			SimpleNode ElseStm = (SimpleNode)n3.jjtGetChild(0).jjtGetChild(4);
			nLabels++;
			processStmList((SimpleNode) ElseStm, j, fich);
			nLabels--;
		}		
		
		nLabel=l;
		int n=nLabel+1;		
		
		fich.write(("\tgoto Label" + n + "\n").getBytes());
		
		fich.write(("Label" + nLabel + ":\n").getBytes());		//label do if
		nLabels++;
		processStmList((SimpleNode) StmLst, j, fich);		//statements do if
		nLabels--;
		
		nLabel=l+1;
		fich.write(("Label" + nLabel + ":\n").getBytes());
		
		if(nLabel > maxLabel)
			maxLabel=nLabel;
		
	}

	/**
	 * Process assign.
	 *
	 * @param SimpleNode the n3
	 * @param Integer the j
	 * @param FileOutputStream the fich
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	private void processAssign(SimpleNode n3, int j, FileOutputStream fich) throws IOException 
	{		
		
		SimpleNode RHS = (SimpleNode)n3.jjtGetChild(0).jjtGetChild(1);
		SimpleNode LHS = (SimpleNode)n3.jjtGetChild(0).jjtGetChild(0);
		
		if(LHS.jjtGetChild(1).toString().equals("ARRAYACCESS"))
		{
			processLHS(LHS, j, fich, 0);
			processRHS(RHS, j, fich);
			fich.write("\tiastore\n".getBytes());
		}
		else
		{
			processRHS(RHS, j, fich);	
			processLHS(LHS, j, fich, 0);
		}
	}
	
	/**
	 * Process rhs.
	 *
	 * @param SimpleNode the rHS
	 * @param Integer the j
	 * @param FileOutputStream the fich
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	void processRHS(SimpleNode RHS, int j, FileOutputStream fich) throws IOException
	{
		SimpleNode FirstTerm = (SimpleNode) RHS.jjtGetChild(0);
		
		if(FirstTerm.toString().equals("TERM"))
			processTerm(FirstTerm, j, fich);
		else
		{
			SimpleNode ArraySize = (SimpleNode) RHS.jjtGetChild(0).jjtGetChild(0);
			
			if(ArraySize.toString().equals("INTEGER"))
				fich.write(("\tbipush " + ArraySize.arraysize + "\n\tnewarray int\n").getBytes());
			else if(ArraySize.toString().equals("ID"))
			{
				int ind = ArraySize.arraysize2.indexOf(".");
				String tmp = ArraySize.arraysize2.substring(0, ind);
				if(variaveisFunc.containsKey(tmp))
				{
					if(this.variaveisFunc.get(tmp) >= 0 && this.variaveisFunc.get(tmp) <=3)
						fich.write(("\tiload_" + this.variaveisFunc.get(tmp) + "\n\tnewarray int\n").getBytes());
					else
						fich.write(("\tiload " + this.variaveisFunc.get(tmp) + "\n\tnewarray int\n").getBytes());
				}
				else
				{
					String s = "\tgetstatic " + nomeModulo + "/" + tmp + " ";
					String tmp1= tabelaSimbolosGlobal.get(tmp);
					
					if(tmp1.equals("Scalar"))
						s += "I\n";
					else
						s +="[I\n";
					
					s+= "\nnewarray int\n";
					fich.write(s.getBytes());
				}
			}			
		}
		
		if(RHS.children.length > 2)
		{
			SimpleNode SecondTerm = (SimpleNode) RHS.jjtGetChild(2);
			processTerm(SecondTerm, j, fich);			
			SimpleNode Oper = (SimpleNode) RHS.jjtGetChild(1);
			
			String s7="";
			if(Oper.operation.equals("+"))
				s7 ="\tiadd\n";
			else if(Oper.operation.equals("-"))
				s7 ="\tisub\n";
			else if(Oper.operation.equals("*"))
				s7 ="\timul\n";
			else if(Oper.operation.equals("/"))
				s7 ="\tidiv\n";
			else if(Oper.operation.equals("&"))
				s7 ="\tiand\n";
			else if(Oper.operation.equals("|"))
				s7 ="\tior\n";
			else if(Oper.operation.equals("^"))
				s7 ="\tixor\n";
			else if(Oper.operation.equals("<<"))
				s7 ="\tishl\n";
			else if(Oper.operation.equals(">>"))
				s7 ="\tishr\n";
			else if(Oper.operation.equals(">>>"))
				s7 ="\tiushr\n";

			fich.write(s7.getBytes());
		}	
		
	}
	
	/**
	 * Process term.
	 *
	 * @param SimpleNode the FirstTerm
	 * @param Integer the j
	 * @param FileOutputStream the fich
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	void processTerm(SimpleNode FirstTerm, int j, FileOutputStream fich) throws IOException
	{
		for(int i=0; i < FirstTerm.children.length; i++)
		{
			SimpleNode n1FirstTerm = (SimpleNode) FirstTerm.jjtGetChild(i);
			String valueid="", tmp="";
			int val;
						
			if(n1FirstTerm.toString().equals("INTEGER"))
			{
				val = ((SimpleNode) n1FirstTerm).Valor;		
				
				if( val >= 0 && val <= 5)
					tmp = "\ticonst_" + val + "\n";
				else
					tmp = "\tldc " + val + "\n";
			}
			else if(n1FirstTerm.toString().equals("ID"))
			{
				valueid = ((SimpleNode) n1FirstTerm).valueid;
				
				if(functions.get(j).var_local.containsKey(valueid)) 
				{
					int numVarLocal = variaveisFunc.get(valueid);
					String s="";
					
					if(functions.get(j).var_local.get(valueid).equals("Scalar"))
					{
						if(numVarLocal >=0 && numVarLocal <= 3)
							s = "\tiload_" + numVarLocal + "\n";
						else
							s = "\tiload " + numVarLocal + "\n";
					}
					else
					{
						if(numVarLocal >=0 && numVarLocal <= 3)
							s = "\taload_" + numVarLocal + "\n";
						else
							s = "\taload " + numVarLocal + "\n";
					}
					fich.write(s.getBytes());
				}
				else if (functions.get(j).args.containsKey(valueid))
				{
					int numVarLocal = variaveisFunc.get(valueid);
					String s="";
					
					if(functions.get(j).args.get(valueid).equals("Scalar"))
					{
						if(numVarLocal >=0 && numVarLocal <= 3)
							s = "\tiload_" + numVarLocal + "\n";
						else
							s = "\tiload " + numVarLocal + "\n";
					}
					else
					{
						if(numVarLocal >=0 && numVarLocal <= 3)
							s = "\taload_" + numVarLocal + "\n";
						else
							s = "\taload " + numVarLocal + "\n";
					}
					fich.write(s.getBytes());
				}
				else if (functions.get(j).var_ret.equals(valueid))
				{
					int numVarLocal = variaveisFunc.get(valueid);
					String s="";
					
					if(functions.get(j).type_return.equals("Scalar"))
					{
						if(numVarLocal >=0 && numVarLocal <= 3)
							s = "\tiload_" + numVarLocal + "\n";
						else
							s = "\tiload " + numVarLocal + "\n";
					}
					else
					{
						if(numVarLocal >=0 && numVarLocal <= 3)
							s = "\taload_" + numVarLocal + "\n";
						else
							s = "\taload " + numVarLocal + "\n";
					}
					fich.write(s.getBytes());
				}
				else if(this.tabelaSimbolosGlobal.containsKey(valueid))
				{
					tmp = "\tgetstatic " + nomeModulo + "/" + valueid + " ";
					String tmp1="";
					if(functions.get(j).var_local.containsKey(valueid))
						tmp1= functions.get(j).var_local.get(valueid);
					else if(functions.get(j).args.containsKey(valueid))
						tmp1= functions.get(j).args.get(valueid);
					else if(tabelaSimbolosGlobal.containsKey(valueid))
						tmp1= tabelaSimbolosGlobal.get(valueid);
					
					if(tmp1.equals("Scalar"))
						tmp += "I";
					else
						tmp +="[I";
					
					tmp+="\n";
				}
			}
			else if(n1FirstTerm.toString().equals("ARRAYACCESS"))
			{
				String s1 = ((SimpleNode)n1FirstTerm.jjtGetChild(0).jjtGetChild(0).jjtGetChild(0)).Indice;
				
				String tmp1="";
				if( s1.startsWith("0")|| s1.startsWith("1")
					|| s1.startsWith("2")|| s1.startsWith("3")
					|| s1.startsWith("4")|| s1.startsWith("5")
					|| s1.startsWith("6")|| s1.startsWith("7")
					|| s1.startsWith("8")|| s1.startsWith("9"))	//variavel nao guardada
				{
					int z=-1;
					try{
						z = Integer.parseInt(s1);
					}
					catch (NumberFormatException e) {z=-1;};
					
					if( z >= 0 && z <= 5)		
						tmp1 = "\ticonst_" + s1 + "\n";
					else
						tmp1 = "\tldc " + s1 + "\n";
				}
				else
				{
					if(this.variaveisFunc.containsKey(s1))
					{
						if(variaveisFunc.get(s1) >= 0 && variaveisFunc.get(s1) <= 3)
							tmp1 = "\tiload_" + variaveisFunc.get(s1);
						else
							tmp1 = "\tiload " + variaveisFunc.get(s1);
					}
					else 
					{
						tmp1 = "\tgetstatic " + nomeModulo + "/" + s1 + " ";	
						String type="";
						if(functions.get(j).var_local.containsKey(s1))
							type = functions.get(j).var_local.get(s1);
						else if(functions.get(j).args.containsKey(s1))
							type = functions.get(j).args.get(s1);
						else if(this.tabelaSimbolosGlobal.containsKey(s1))
							type = tabelaSimbolosGlobal.get(s1);
						else if(functions.get(j).var_ret.equals(s1))
							type = functions.get(j).type_return;
						
						if(type.equals("Scalar"))
							tmp1 += "I";
						else
							tmp1 +="[I";
					}
					tmp1+="\n";
				}
				
				fich.write(tmp1.getBytes());
				fich.write("\tiaload\n".getBytes());
				
			}
			else if(n1FirstTerm.toString().equals("SCALARACCESS"))
			{
				if(((SimpleNode)n1FirstTerm).jjtGetNumChildren() != 0)
				{
					fich.write("	arraylength\n".getBytes());
				}
				
			}
			else if(n1FirstTerm.toString().equals("CALL"))
				processCall((SimpleNode) n1FirstTerm.jjtGetParent(), j,fich );
			
			fich.write(tmp.getBytes());
		}
	}
	
	/**
	 * Process lhs.
	 *
	 * @param SimpleNode the lHS
	 * @param Integer the j
	 * @param FileOutputStream the fich
	 * @param Integer the ifid
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	void processLHS(SimpleNode LHS, int j, FileOutputStream fich, int ifid) throws IOException
	{
		String VarEsq = ((SimpleNode)LHS.jjtGetChild(0)).VarEsq;
		
		if(LHS.jjtGetChild(1).toString().equals("ARRAYACCESS"))
		{
			String s1 = ((SimpleNode)LHS.jjtGetChild(1).jjtGetChild(0).jjtGetChild(0).jjtGetChild(0)).Indice;
			
			//carregar var_array
			String s="";		
			
			if(this.tabelaSimbolosGlobal.containsKey(VarEsq))
			{
				s = "\tgetstatic " + nomeModulo + "/" + VarEsq + " ";
				String tmp1= tabelaSimbolosGlobal.get(VarEsq);
				
				if(tmp1.equals("Scalar"))
					s += "I\n";
				else
					s +="[I\n";
				fich.write(s.getBytes());
			}
			else
			{
				int numVarLocal = this.variaveisFunc.get(VarEsq);
				if(numVarLocal >= 0 && numVarLocal <= 3)
					s = "\taload_" + numVarLocal + "\n";	
				else
					s = "\taload " + numVarLocal + "\n";
				fich.write(s.getBytes());
			}
			//carregar indice
			
			String tmp="";
			if( s1.startsWith("0")|| s1.startsWith("1")
				|| s1.startsWith("2")|| s1.startsWith("3")
				|| s1.startsWith("4")|| s1.startsWith("5")
				|| s1.startsWith("6")|| s1.startsWith("7")
				|| s1.startsWith("8")|| s1.startsWith("9"))	//variavel nao guardada
			{
				int i=-1;
				try{
					i = Integer.parseInt(s1);
				}
				catch (NumberFormatException e) {i=-1;};
				
				if( i >= 0 && i <= 5)				
					tmp = "\ticonst_" + s1 + "\n";
				else
					tmp = "\tldc " + s1 + "\n";
			}
			else
			{
				if(this.variaveisFunc.containsKey(s1))
				{
					String type="";
					if(functions.get(j).var_local.containsKey(s1))
						type = functions.get(j).var_local.get(s1);
					else if(functions.get(j).args.containsKey(s1))
						type = functions.get(j).args.get(s1);
					else if(this.tabelaSimbolosGlobal.containsKey(s1))
						type = tabelaSimbolosGlobal.get(s1);
					else if(functions.get(j).var_ret.equals(s1))
						type = functions.get(j).type_return;
					
					if(type.equals("Scalar"))
					{
						if(variaveisFunc.get(s1) >= 0 && variaveisFunc.get(s1) <= 3)
							tmp = "\tiload_" + variaveisFunc.get(s1);
						else
							tmp = "\tiload " + variaveisFunc.get(s1);
					}
					else
					{
						if(variaveisFunc.get(s1) >= 0 && variaveisFunc.get(s1) <= 3)
							tmp = "\taload_" + variaveisFunc.get(s1);
						else
							tmp = "\taload " + variaveisFunc.get(s1);
					}
				}
				else 
				{
					tmp = "getstatic " + nomeModulo + "/" + s1 + " ";	
					String type="";
					if(functions.get(j).var_local.containsKey(s1))
						type = functions.get(j).var_local.get(s1);
					else if(functions.get(j).args.containsKey(s1))
						type = functions.get(j).args.get(s1);
					else if(this.tabelaSimbolosGlobal.containsKey(s1))
						type = tabelaSimbolosGlobal.get(s1);
					else if(functions.get(j).var_ret.equals(s1))
						type = functions.get(j).type_return;
					
					if(type.equals("Scalar"))
						tmp += "I";
					else
						tmp +="[I";
				}
				tmp+="\n";
			}
			
			fich.write(tmp.getBytes());
		}
		else
		{
			if(functions.get(j).var_local.containsKey(VarEsq) || functions.get(j).args.containsKey(VarEsq) || functions.get(j).var_ret.equals(VarEsq))
			{
				if(!variaveisFunc.containsKey(VarEsq))	
					variaveisFunc.put(VarEsq, variaveisFunc.size());
				
				int numVarLocal = variaveisFunc.get(VarEsq);
				String  s, tipo="";
				if(functions.get(j).var_local.containsKey(VarEsq))
					tipo = functions.get(j).var_local.get(VarEsq);
				else if(functions.get(j).args.containsKey(VarEsq))
					tipo = functions.get(j).args.get(VarEsq);
				else if(this.tabelaSimbolosGlobal.containsKey(VarEsq))
					tipo = tabelaSimbolosGlobal.get(VarEsq);
				else if(functions.get(j).var_ret.equals(VarEsq))
					tipo = functions.get(j).type_return;
				if(ifid==0)
				{
					if(tipo.equals("Scalar"))
					{
						if(numVarLocal >= 0 && numVarLocal <= 3)
							s = "\tistore_" + numVarLocal + "\n";
						else
							s = "\tistore " + numVarLocal + "\n";
					}
					else
					{
						if(numVarLocal >= 0 && numVarLocal <= 3)
							s = "\tastore_" + numVarLocal + "\n";
						else
							s = "\tastore " + numVarLocal + "\n";
					}
				}
				else
				{
					if(tipo.equals("Scalar"))
					{
						if(numVarLocal >= 0 && numVarLocal <= 3)
							s = "\tiload_" + numVarLocal + "\n";
						else
							s = "\tiload " + numVarLocal + "\n";
					}
					else
					{
						if(numVarLocal >= 0 && numVarLocal <= 3)	
							s = "\taload_" + numVarLocal + "\n";
						else
							s = "\taload " + numVarLocal + "\n";
					}
				}
				fich.write(s.getBytes());
			}
			else if(this.tabelaSimbolosGlobal.containsKey(VarEsq))
			{
				String tmp="";
				if(ifid==0)
					tmp = "\tputstatic " + nomeModulo + "/" + VarEsq + " ";
				else
					tmp = "\tgetstatic " + nomeModulo + "/" + VarEsq + " ";
					
				String tmp1= tabelaSimbolosGlobal.get(VarEsq);
				
				if(tmp1.equals("Scalar"))
					tmp += "I";
				else
					tmp +="[I";
				
				tmp+="\n";
				fich.write(tmp.getBytes());
			}
			
			
			if(LHS.jjtGetChild(1).toString().equals("SCALARACCESS"))
				if(((SimpleNode)LHS.jjtGetChild(1)).jjtGetNumChildren() != 0)
					fich.write("	arraylength\n".getBytes());										 
		}		
	}

	/**
	 * Process call.
	 *
	 * @param SimpleNode the n3
	 * @param Integer the j
	 * @param FileOutputStream the fich
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	void processCall(SimpleNode n3, int j, FileOutputStream fich) throws IOException
	{
		String CalledFunc="",s4="", s5="";
		s4=((SimpleNode)n3.jjtGetChild(0).jjtGetChild(0)).CallFunc;
		
		if(((SimpleNode)n3.jjtGetChild(0)).children.length != 1)
			s5=((SimpleNode)n3.jjtGetChild(0).jjtGetChild(1)).CallFunc;
		
		CalledFunc=s5;
		try{
			if(s5.equals(""))
				CalledFunc=s4;
		}catch (java.lang.NullPointerException e)
		{
			CalledFunc=s4;
		}	
		boolean flag1=false;
		if(CalledFunc.startsWith(this.nomeModulo+"/"))
		{
			int ind1 = CalledFunc.indexOf("/");
			CalledFunc = CalledFunc.substring(ind1+1,CalledFunc.length());
			flag1=true;
		}
		
		ArrayList<String> args = new ArrayList<String>();
		SimpleNode argList = null;
		
		if(((SimpleNode)n3.jjtGetChild(0)).children.length >= 2 )
		{
			boolean flag=false;
			for(int i=0; i < functions.size(); i++)
				if(functions.get(i).name.equals(CalledFunc))
					flag=true;
			
			SimpleNode n5;	
			if(flag && !flag1)
				n5 = ((SimpleNode)n3.jjtGetChild(0).jjtGetChild(1));	//argumentList
			else
			{
				if(((SimpleNode)n3.jjtGetChild(0)).children.length == 3)
					n5 = ((SimpleNode)n3.jjtGetChild(0).jjtGetChild(2));	//argumentList
				else
					n5 = ((SimpleNode)n3.jjtGetChild(0).jjtGetChild(1));	//argumentList
			}
			argList=n5;
			
			
			if(n5.jjtGetNumChildren() > 0)
			{
				String arg = ((SimpleNode)n5.jjtGetChild(0).jjtGetChild(0)).calledArg;
				args.add(arg);
				
				for(int v=1; v < n5.children.length; v++)
				{
					SimpleNode n6 = (SimpleNode) n5.children[v];
					for( int z=0; z < ((SimpleNode) n6.jjtGetChild(0)).children.length; z++)												
						args.add(((SimpleNode)((SimpleNode) n6.jjtGetChild(0)).children[z]).calledArg);
				}
				
				for(int y=0; y < args.size(); y++)
				{
					String s = "",tmp=""; s+= '"';
					if(args.get(y).contains(s) 
							|| args.get(y).startsWith("0")|| args.get(y).startsWith("1")
							|| args.get(y).startsWith("2")|| args.get(y).startsWith("3")
							|| args.get(y).startsWith("4")|| args.get(y).startsWith("5")
							|| args.get(y).startsWith("6")|| args.get(y).startsWith("7")
							|| args.get(y).startsWith("8")|| args.get(y).startsWith("9"))	//variavel nao guardada
					{
						int i=-1;
						try{
							i = Integer.parseInt(args.get(y));
						}
						catch (NumberFormatException e) {i=-1;};
						
						if( i >= 0 && i <= 5)
							tmp = "\ticonst_" + args.get(y) + "\n";
						else
							tmp = "\tldc " + args.get(y) + "\n";
					}
					else
					{
						if(this.variaveisFunc.containsKey(args.get(y)))
						{
							String type="";
							if(functions.get(j).var_local.containsKey(args.get(y)))
								type = functions.get(j).var_local.get(args.get(y));
							else if(functions.get(j).args.containsKey(args.get(y)))
								type = functions.get(j).args.get(args.get(y));
							else if(this.tabelaSimbolosGlobal.containsKey(args.get(y)))
								type = tabelaSimbolosGlobal.get(args.get(y));
							else if(functions.get(j).var_ret.equals(args.get(y)))
								type = functions.get(j).type_return;
							
							if(type.equals("Scalar"))
							{
								if(variaveisFunc.get(args.get(y)) >= 0 && variaveisFunc.get(args.get(y)) <= 3)
									tmp = "\tiload_" + variaveisFunc.get(args.get(y));
								else
									tmp = "\tiload " + variaveisFunc.get(args.get(y));
							}
							else
							{
								if(variaveisFunc.get(args.get(y)) >= 0 && variaveisFunc.get(args.get(y)) <= 3)
									tmp = "\taload_" + variaveisFunc.get(args.get(y));
								else
									tmp = "\taload " + variaveisFunc.get(args.get(y));
							}
						}
						else 
						{
							tmp = "\tgetstatic " + nomeModulo + "/" + args.get(y) + " ";	
							String type="";
							if(functions.get(j).var_local.containsKey(args.get(y)))
								type = functions.get(j).var_local.get(args.get(y));
							else if(functions.get(j).args.containsKey(args.get(y)))
								type = functions.get(j).args.get(args.get(y));
							else if(this.tabelaSimbolosGlobal.containsKey(args.get(y)))
								type = tabelaSimbolosGlobal.get(args.get(y));
							else if(functions.get(j).var_ret.equals(args.get(y)))
								type = functions.get(j).type_return;
							
							if(type.equals("Scalar"))
								tmp += "I";
							else
								tmp +="[I";
						}
						tmp+="\n";
					}
					
					fich.write(tmp.getBytes());
				}
			}
		}
		
		
		String s7="", s8="";
		boolean flag=false;
		for(int g=0; g < functions.size(); g++)
		{
			if(functions.get(g).name.equals(CalledFunc))
			{
				s7 = "\tinvokestatic " + this.nomeModulo + "/" + CalledFunc + "(";
				Object[] keys = functions.get(g).args.keySet().toArray();
				for(int h=0; h < keys.length; h++)
				{
					if(functions.get(g).args.get(keys[h]).equals("Scalar"))
						s7 += "I";
					else if(functions.get(g).args.get(keys[h]).equals("ScalarArray"))
						s7 += "[I";
					else
					{
						s7 += "Ljava/lang/String"; 
						if(h!=keys.length-1)
							s7 += ";";
					}
				}
				flag=true;
				
				if(functions.get(g).type_return.equals("void"))
					s7+= ")V";
				else if(functions.get(g).type_return.equals("Scalar"))
					s7+= ")I";
				else if(functions.get(g).type_return.equals("ScalarArray"))
					s7+= ")[I";
				
				s7+="\n";
				fich.write(s7.getBytes());
			}
		}		
		
		if(!flag)
		{			
			s8 = "\tinvokestatic " + CalledFunc + "(";
			
			if(CalledFunc.equals("io/print") || CalledFunc.equals("io/println"))
			{
				if( argList.jjtGetNumChildren() == 2)
					s8+="Ljava/lang/String;I";
				if( argList.jjtGetNumChildren() == 1)
				{
					String tmp3=((SimpleNode)((SimpleNode) argList.jjtGetChild(0)).jjtGetChild(0)).calledArg;
					String tmp4=""; tmp4+='"';
					if( tmp3.startsWith(tmp4))
							s8+="Ljava/lang/String;";
					else
							s8+="I";
				}
			}
						
			s8+=")V\n";
			fich.write(s8.getBytes());			
		}		
	}
		
	/**
	 * Constroi globais.
	 *
	 * @param SimpleNode the n
	 * @param FileOutputStream the fich
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	void constroiGlobais(SimpleNode n,FileOutputStream fich) throws IOException
	{
			ArrayList<String> arraysInicializados = new ArrayList<String>();
			
			for(int i=0; i < n.children.length; i++)
			{
				SimpleNode n1 = (SimpleNode) n.children[i];
				if(n.children[i].toString().equals("ID")) 
				{
					String n2 = ".class public "+ n1.moduleName + "\n";
					nomeModulo=n1.moduleName;
					fich.write(n2.getBytes());
					String n3 = ".super java/lang/Object"+ "\n";
					fich.write(n3.getBytes());
				}
				else if(n.children[i].toString().equals("DECLARATION")) //variaveis globais
				{
					String nomeVar="", arraysize2="";
					int arraysize=0, Initialvalue=0;
					for(int j=0; j < n1.jjtGetNumChildren(); j++)
					{
						if(n1.jjtGetChild(j).toString().equals("VAR"))
						{
							SimpleNode n2 = (SimpleNode) n1.jjtGetChild(j).jjtGetChild(0);
							nomeVar=n2.varGlobal;
						}
						else if(n1.jjtGetChild(j).toString().equals("ARRAYSIZE"))
						{
							SimpleNode n2 = (SimpleNode) n1.jjtGetChild(j).jjtGetChild(0);
							if(n2.arraysize != 0)
							{
								arraysize=n2.arraysize;
								GlobalVariables gv1 = new GlobalVariables(nomeVar, tabelaSimbolosGlobal.get(nomeVar), arraysize);
								this.globais.add(gv1);
							}
							else
								arraysize2=((SimpleNode) n1.jjtGetChild(j).jjtGetChild(0)).arraysize2;
						}
						else if(n1.jjtGetChild(j).toString().equals("INTEGER"))
						{
							SimpleNode n2 = (SimpleNode) n1.jjtGetChild(j);
							Initialvalue=n2.valorInicial;
						}	
					}
					String n2 = ".field static "+ nomeVar;
					String type = this.tabelaSimbolosGlobal.get(nomeVar);
					if(type.equals("Scalar"))	
					{
						n2 += " I ";		
						GlobalVariables gv1 = new GlobalVariables(nomeVar, tabelaSimbolosGlobal.get(nomeVar), Initialvalue);
						this.globais.add(gv1);
						
						if(Initialvalue != 0)
							n2+= "= " + Initialvalue;
						fich.write((n2+"\n").getBytes());
					}
					else
					{
						n2 += " [I";
						fich.write((n2+"\n").getBytes());
						if(arraysize != 0)
						{
							String n4 = ".method static public <clinit>()V\n.limit stack 2\n.limit locals 0\n\tbipush ";
							n4 += arraysize + "\n\tnewarray int\nputstatic " + nomeModulo + "/" + nomeVar +" [I\nreturn\n.end method\n";
							arraysInicializados.add(n4);
						}
						else 	//tamanho do array e' dependente de outra variavel
						{
							int pos = arraysize2.indexOf(".");
							
							if(pos != -1)
							{
								String s1 = arraysize2.substring(0,pos);
								
								int val = 0;
								for(int g=0; g < globais.size(); g++)
								{
									if(globais.get(g).VarName.equals(s1))
									{
										val=globais.get(g).CurrentValue;
									}
								}
								
								String n4 = ".method static public <clinit>()V\n.limit stack 2\n.limit locals 0\n\tbipush ";
								n4 += val + "\n\tnewarray int\nputstatic " + nomeModulo + "/" + nomeVar +" [I\nreturn\n.end method\n";
								arraysInicializados.add(n4);
								
								GlobalVariables gv1 = new GlobalVariables(nomeVar, tabelaSimbolosGlobal.get(nomeVar), val);
								this.globais.add(gv1);
							}
						}
					}					
				}
			}
			
			for(int g=0; g < arraysInicializados.size(); g++)
				fich.write((arraysInicializados.get(g)+"\n").getBytes());
	}	
}
